/* Copyright 2007 Noah B. Jacobsen.
 * 
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <time.h>

#include "alloc.h"
#include "blockio.h"
#include "open.h"
#include "mod2sparse.h"
#include "mod2dense.h"
#include "mod2convert.h"
#include "ncode.h"
#include "rand.h"
#include "enc.h"
#include "dec.h"
#include "biawgn_harq.h"

time_t time0;
int ewords, rwords;

int biawgn_harq(double snr_dB, int words, int gflag, 
	   double *uer, double *cer, double *wer)
{  
  int K = N-M;
  double R = (double)K/N;
  double snr = pow(10,snr_dB/10);
  double nvar = (double)1/2/R/snr;
  //double nvar = (double)1/R/snr;
  double std_dev = sqrt(nvar);
  char info_bits[K], code_bits[N];
  double awgn_noise, recvd, lratio[N];
  char dec_bits[N], pchk[M], chks[M];
  double bitpr[N];

  int invalid,block_no,iters,k;
  int u_errs, c_errs;
  double atime, rtime;
  time_t time1;

  //fprintf(stdout,"\n     std_dev = %f\n\n",std_dev);
  /* Allocate space for encoder */
  mod2dense *w, *v, **codeWords, **harqBits;
  
  if( gflag ){
    w = mod2dense_allocate(K,1);
    v = mod2dense_allocate(codeLength[0]-K,1);

    codeWords = chk_alloc(harqNum+1, sizeof *codeWords);
    if(harqNum>0) harqBits = chk_alloc(harqNum, sizeof *harqBits);

    for(k=0;k<=harqNum;k++){
      codeWords[k] = mod2dense_allocate(codeLength[k],1);
      if(k>0) harqBits[k-1] = mod2dense_allocate(codeLength[k]-codeLength[k-1],1);
    }
  }

  /* Loop over codewords */
  for( block_no = 0; block_no < words; block_no++ ){

    if( gflag ){
      rand_bits(info_bits,K);
      
      /* Compute encoded block. */
      harq_encode (info_bits, code_bits, w, v, codeWords, harqBits);

      /* Check that encoded block is a code word. */

      mod2sparse_mulvec (H, code_bits, chks);
    
      for (k = 0; k<M; k++){
	if (chks[k]==1){
	  fprintf(stdout,"\nOutput block %d is not a code word!  (Fails check %d)\n",block_no,k);
	  exit(1);
	}
      }
    }

    else
      for(k = 0; k<N; k++)
	code_bits[k] = 0;


    /* Modulate BPSK and add noise. */

    for(k = 0; k<N; k++){
      awgn_noise = std_dev * rand_gaussian();
      recvd = pow(-1,code_bits[k]) + awgn_noise;

      lratio[k] = exp(2*recvd/nvar);
    }

    /* Call decoder. */
    iters = prprp_decode (H, lratio, dec_bits, pchk, bitpr);
    
    /* Compute error rates. */

    for( k = 0, u_errs = 0; k<K; k++ )
      if(gflag)
	u_errs += dec_bits[codeLength[0]-cols[k+(codeLength[0]-K)]-1]!=info_bits[k];
      else
	u_errs += dec_bits[k]!=code_bits[k];

    for( k = 0, c_errs = 0; k<N; k++ )
      c_errs += dec_bits[k]!=code_bits[k];
    
    uer[0] = (uer[0]*block_no + (double)u_errs/K)/(block_no+1);
    cer[0] = (cer[0]*block_no + (double)c_errs/N)/(block_no+1);
    wer[0] = (wer[0]*block_no + (double)(c_errs>0))/(block_no+1);

    /* Report progress. */
    
    ewords += 1;
    rwords -= 1;

    if( (block_no+1)%Report==0 ){
      time(&time1);
      atime = (double)difftime(time1,time0)/ewords;
      rtime = (double)rwords*atime/60;

      fprintf(stdout,"%5d: uer = %1.2e, cer = %1.2e, wer = %1.2e (rtime = %.2f)\n",block_no,uer[0],cer[0],wer[0],rtime);
      fflush(stdout);
    }
  }

  if( gflag ){
    mod2dense_free(w);
    mod2dense_free(v);

    for(k=0;k<=harqNum;k++){
      mod2dense_free(codeWords[k]);
      if(k>0) mod2dense_free(harqBits[k-1]);
    }
    
    free(codeWords);
    if(harqNum>0) free(harqBits);
  }

  return 0;
}


int main(int argc, char **argv){

  char *pchk_file, *gen_file, *out_file;
  char **arg_pr, *instr, junk;
  FILE *of;
  int words[10], seed;
  double snr_dB[10];
  double *uer, *cer, *wer;
  int L = 1, L1 = 1, gflag = 1;
  int len, l;

  /* Read input arguments. */

  if( !(pchk_file = argv[1]) || !(gen_file = argv[2]) )
    usage();
  
  len = strlen(gen_file);

  if( strcmp(gen_file+len-4,".gen")!=0 ){
    if( sscanf(gen_file,"%lf",snr_dB,&junk) != 1 )
      usage();
      
    gflag = 0;
    arg_pr = argv+2;
  }
  else
    arg_pr = argv+3;

  if( (instr = arg_pr[0]) && (instr[0] != '\0') )
    while( sscanf(instr,"%lf %n%c",snr_dB+L-1,&l,&junk) == 2 ){ 
      L += 1;
      instr += l;
    }
  else{
    fprintf(stdout,"\nSNR didn't parse.\n");
    usage();
  }
  
  if( (instr = arg_pr[1]) && (instr[0] != '\0') )
    while( sscanf(instr,"%d %n%c",words+L1-1,&l,&junk) == 2 ){ 
      L1 += 1;
      instr += l;
     }
  else{
    fprintf(stdout,"\nWORDS didn't parse.\n");
    usage();
  }
  
  if( L != L1 ){
    fprintf(stdout,"\nSNR points does not equal WORDS length.\n");
    usage();
  }
  
  if( !arg_pr[2] || sscanf(arg_pr[2],"%d%c",&seed,&junk)!=1 
      || !arg_pr[3] || sscanf(arg_pr[3],"%d%c",&max_iter,&junk)!=1  
      || !(out_file = arg_pr[4]) )
    usage();


  /* Echo input arguments. */
  system("echo $HOSTNAME:");
  fprintf(stdout,"%s ",argv[0]);
  for( l=1; l<argc; l++ ){
    if( gflag&&(l==3 || l==4) || (!gflag)&&(l==2 || l==3) )
      fprintf(stdout,"\"%s\" ",argv[l]);
    else
      fprintf(stdout,"%s ",argv[l]);
  }
  fprintf(stdout,"\n\nParity matrix: %s",pchk_file);
  if(gflag)
    fprintf(stdout,"\nGenerator matrix: %s",gen_file);
  else
    fprintf(stdout,"\nNo generator matrix specified.");
  fprintf(stdout,"\nOutput file: %s\nSNR = %1.2f dB (%d words)",out_file,snr_dB[0],words[0]);
  for (l = 1; l<L; l++)
    fprintf(stdout,", %1.2f dB (%d words)",snr_dB[l],words[l]);
  fprintf(stdout,"\nSEED = %d",seed);
  fprintf(stdout,"\nITERATIONS = %d\n",max_iter);
  fflush(stdout);

  /* Open parity check file. */

  read_pchk(pchk_file);

  /* Read generator matrix file if specified. */

  if( gflag )
    read_gen(gen_file);

  /* Set random seed. */

  rand_seed(10*seed+3);
  
  /* Call main program. */
  
  uer = chk_alloc (L, sizeof *uer);
  cer = chk_alloc (L, sizeof *cer);
  wer = chk_alloc (L, sizeof *wer);

  for(l = 0, rwords = 0; l<L; l++)
    rwords += words[l]; 
  ewords = 0;

  time( &time0);
  for(l = 0; l<L; l++){
    fprintf(stdout,"\n--> SNR = %1.2f, WORDS = %d <--\n",snr_dB[l],words[l]);
    fflush(stdout);
    biawgn_harq(snr_dB[l],words[l],gflag,uer+l,cer+l,wer+l);

    fprintf(stdout,"\nSNR = %1.2f dB: UER = %1.3e, CER = %1.3e, WER = %1.3e\n",snr_dB[l],uer[l],cer[l],wer[l]);
  }


  /* Open output file. */

  of = open_file_std(out_file,"w");
  if (of==NULL){ 
    fprintf(stderr,"Can't create file for output data: %s\n",of);
    exit(1);
  }

  /* Write output. */
  
  for(l = 0; l<L; l++){
    fprintf(of,"snr(%d)=%1.3f; uer(%d)=%1.4e; cer(%d)=%1.4e; wer(%d)=%1.4e;\n",l+1,snr_dB[l],l+1,uer[l],l+1,cer[l],l+1,wer[l]);    
  }

  /* Close output file. */

  if( ferror(of) || fclose(of)!=0 ){ 
    fprintf(stderr,"Error writing output data to %s\n",out_file);
    exit(1);
  }

  free(uer);
  free(cer);
  free(wer);
 
  return 0;
}

/* PRINT USAGE MESSAGE AND EXIT. */

void usage(void){ 
  fprintf(stderr,
	  "Usage:   biawgn_harq pchk-file [gen-file] SNR WORDS SEED ITERATIONS out-file\n");
  exit(1);
}


/*
 *    EXAMPLE USAGE
 *
 *    biawgn_harq baseCodeName[harqNum].pchk baseCodeName.gen "-.3 -.1" "4000 8000" 90310 -100 outfile.m
 *
 *    or for all zeros codeword:
 *    biawgn_harq baseCodeName[harqNum].pchk "-.3 -.1" "4000 8000" 90310 -100 outfile.m
 *
 *
 */
