/* Copyright 2007 Noah B. Jacobsen.
 * 
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

/******************************************************************************/
/* *******Includes code from:*********                                        */
/* Copyright (c) 2000, 2001 by Radford M. Neal                                */
/*                                                                            */
/* Permission is granted for anyone to copy, use, modify, or distribute this  */
/* program and accompanying programs and documents for any purpose, provided  */
/* this copyright notice is retained and prominently displayed, along with    */
/* a note saying that the original programs are available from Radford Neal's */
/* web page, and note is made of any changes made to the programs.            */
/******************************************************************************/
#include <stdio.h>
#include <stdlib.h>
#include <math.h>

#include "alloc.h"
#include "mod2sparse.h"
#include "mod2dense.h"
#include "mod2convert.h"
#include "ncode.h"
#include "check.h"
#include "dec.h"

/* GLOBAL VARIABLES.  Declared in dec.h. */

int block_no;	/* Number of current block, from zero */
int max_iter;	/* Maximum number of iteratons of decoding to do */

/* DECODE USING PROBABILITY PROPAGATION.  */

int prprp_decode
( mod2sparse *H,	/* Parity check matrix */
  double *lratio,	/* Likelihood ratios for bits */
  char *dblk,		/* Place to store decoding */
  char *pchk,		/* Place to store parity checks */
  double *bprb		/* Place to store bit probabilities */
)
{ 
  int N, n, c;

  N = mod2sparse_cols(H);

  /* Initialize probability and likelihood ratios, and find initial guess. */

  initprp(H,lratio,dblk,bprb);

  /* Do up to abs(max_iter) iterations of probability propagation, stopping
     early if a codeword is found, unless max_iter is negative. */

  for (n = 0; ; n++)
  { 
    c = check(H,dblk,pchk);
    
    if (n==max_iter || n==-max_iter || (max_iter>0 && c==0))
    { break; 
    }

    iterprp(H,lratio,dblk,bprb);
  }
  return n;
}


/* INITIALIZE PROBABILITY PROPAGATION.  Stores initial ratios, probabilities,
   and guess at decoding. */

void initprp
( mod2sparse *H,	/* Parity check matrix */
  double *lratio,	/* Likelihood ratios for bits */
  char *dblk,		/* Place to store decoding */
  double *bprb		/* Place to store bit probabilities, 0 if not wanted */
)
{ 
  mod2entry *e;
  int N,j;

  N = mod2sparse_cols(H);

  for (j = 0; j<N; j++){ 
    for (e = mod2sparse_first_in_col(H,j);
	 !mod2sparse_at_end(e);
	 e = mod2sparse_next_in_col(e)){ 
      e->pr = lratio[j];  /* Initialize bit llrs */
      e->lr = 1;          /* Initialize check llrs */
    }
    if (bprb) bprb[j] = lratio[j];
    dblk[j] = lratio[j]<1;
  }
}


/* DO ONE ITERATION OF PROBABILITY PROPAGATION. */

void iterprp
( mod2sparse *H,	/* Parity check matrix */
  double *lratio,	/* Likelihood ratios for bits */
  char *dblk,		/* Place to store decoding */
  double *bprb		/* Place to store bit probabilities, 0 if not wanted */
  )
{
  double pr, dl, t;
  mod2entry *e;
  int N, M;
  int i, j;
  int num_nan = 0, num_inf = 0;
  M = mod2sparse_rows(H);
  N = mod2sparse_cols(H);
  
  /* Compute check llrs */
  
  for (i = 0; i<M; i++){ 
    dl = 1;
    for (e = mod2sparse_first_in_row(H,i);
         !mod2sparse_at_end(e);
         e = mod2sparse_next_in_row(e)){ 
      
      e->lr = dl;
      dl *= 1 - 2/(1+e->pr);

    }

    dl = 1;
    for (e = mod2sparse_last_in_row(H,i);
         !mod2sparse_at_end(e);
         e = mod2sparse_prev_in_row(e)){ 

      t = e->lr * dl;

      e->lr = (1+t)/(1-t);
      dl *= 1 - 2/(1+e->pr);
      
     /*  if ( isinf(dl) ) */
/* 	num_inf += 1; */
    }
  }

  /* Compute bit llrs, codeword estimate */

  for (j = 0; j<N; j++){
    pr = lratio[j];
    for (e = mod2sparse_first_in_col(H,j);
         !mod2sparse_at_end(e);
         e = mod2sparse_next_in_col(e)){ 

      e->pr = pr;
      pr *= e->lr;
    }

    if ( isnan(pr) ){
      pr = 1;
      num_nan += 1;
    }
/*     else if ( isinf(pr) ) */
/*       num_inf += 1; */

    if (bprb) bprb[j] = pr;
    dblk[j] = pr<1;

    pr = 1;
    for (e = mod2sparse_last_in_col(H,j);
         !mod2sparse_at_end(e);
         e = mod2sparse_prev_in_col(e)){ 
 
      e->pr *= pr;
      pr *= e->lr;
      
      if ( isnan(e->pr) ){
	e->pr = 1;
	num_nan += 1;
      }
      else if ( isinf(pr) )
	num_inf += 1;
    }
  }
  
  /*if (num_nan>0)
    fprintf(stdout,"\t\t%d nans\n",num_nan);*/

 /*  if (num_inf>0) */
/*     fprintf(stdout,"\t\t%d infs\n",num_inf); */
}


