/* Copyright 2007 Noah B. Jacobsen.
 * 
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
/* MAKE-GEN.C - Make generator matrices from rate-compatible LDPC codes */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>

#include "alloc.h"
#include "intio.h"
#include "open.h"
#include "mod2sparse.h"
#include "mod2dense.h"
#include "mod2convert.h"
#include "ncode.h"

void makeBaseCodeGenerator(FILE *);
void makeExtCodeGenerator(FILE *,int,int);
void usage(void);


int main
( int argc,
  char **argv
)
{
  char *baseCodeName, codeName[100], extCodeName[100], genFile[100];
  int n,numCodes,N0,M0;
  char junk;
  FILE *f;

  /* Look at arguments. */

  if ( !argv[1] || sscanf(argv[1],"%d%c",&numCodes,&junk)!=1 || numCodes<=0
       || !(baseCodeName = argv[2]) || argc!=3 )
    { usage();
    }
  
  /* Read daughter code parity matrix. */

  strcat(strcpy(codeName,baseCodeName),"0.pchk");
  read_pchk(codeName);

  if (N<=M)
  { fprintf(stderr,
     "Can't encode if number of bits (%d) isn't greater than number of checks (%d)\n",N,M);
    exit(1);
  }

  /* Create generator matrix file. */

  strcat(strcpy(genFile,baseCodeName),".gen");
  f = open_file_std(genFile,"wb");

  if (f==NULL)
  { fprintf(stderr,"Can't create generator matrix file: %s\n",genFile);
    exit(1);
  }
  intio_write(f,('G'<<8)+0x80);

  //write number of codes
  intio_write(f,numCodes);

  /* Allocate space for row and column permutations of base matrix. */

  cols = chk_alloc (N, sizeof *cols);
  rows = chk_alloc (M, sizeof *rows);

  /* Create base code generator matrix. */

  fprintf(stderr,"Computing generator matrix: 0/%d.",numCodes-1);
  makeBaseCodeGenerator(f);

  /* Create extension code generator matrices. */

  for(n=1; n<numCodes; n++){
    N0=N; 
    M0=M;
    mod2sparse_free(H);
    
    fprintf(stderr,"Computing generator matrix: %d/%d.",n,numCodes-1);

    sprintf(codeName,"%s%d.pchk",baseCodeName,n);
    read_pchk(codeName);
    
    makeExtCodeGenerator(f,N0,M0);
  }

  /* Check for error writing file. */

  if( ferror(f) || fclose(f)!=0 )
  { fprintf(stderr,"Error writing to generator matrix file\n");
    exit(1);
  }

  return 0;
}


void makeExtCodeGenerator(FILE *f, int N0, int M0){

  mod2dense *DH, *P0, *P, *L0, *L, *LI, *G1;
  int *row1, *col1;
  int numNewSymbols,c,i,j,k;
  int *new_bit_index;
  int bit;
  
  if(M-M0!=N-N0){
    fprintf(stderr,"Dimension mismatch in developing extension code generator matrix\n");
    exit(1);
  }
  numNewSymbols = N-N0;

  DH = mod2dense_allocate(M,N);
  P0 = mod2dense_allocate(M,N0);
  P  = mod2dense_allocate(numNewSymbols,N0);
  L0 = mod2dense_allocate(M,numNewSymbols);
  L  = mod2dense_allocate(numNewSymbols,numNewSymbols);
  LI = mod2dense_allocate(numNewSymbols,numNewSymbols);
  G1 = mod2dense_allocate(numNewSymbols,N0);

  new_bit_index = chk_alloc(numNewSymbols, sizeof *new_bit_index);
  mod2sparse_to_dense(H,DH);

  /* Extract P sub-matrix */

  col1 = chk_alloc (N0, sizeof *col1);
  for(j=0;j<N0;j++) col1[j]=j;

  row1 = chk_alloc (numNewSymbols, sizeof *row1);
  for(j=M0;j<M;j++) row1[j-M0]=j;

  mod2dense_copycols(DH,P0,col1);
  mod2dense_copyrows(P0,P,row1);
  mod2dense_free(P0);
  
  /* Extract L sub-matrix */

  free(col1);
  col1 = chk_alloc (numNewSymbols, sizeof *col1);
  for(j=N0;j<N;j++) col1[j-N0]=j;

  mod2dense_copycols(DH,L0,col1);
  mod2dense_copyrows(L0,L,row1);
  mod2dense_free(L0);
  

  fprintf(stderr," N=%d, M=%d, N0=%d, M0=%d. ",N,M,N0,M0);

  int NUMELIMONES = 0;

  /* Gauss. elim. on P to map L->I */
  for( j=numNewSymbols-1; j>=0; j--){
    for( i=0; i<numNewSymbols; i++){
      if( i!=j){
	if( mod2dense_get(L,i,j)==1){
	  
	  for( k=0; k<N0; k++){
	    bit = (mod2dense_get(P,j,k) + mod2dense_get(P,i,k)) %2;
	    
	    //fprintf(stderr,"%d + %d = %d\n",mod2dense_get(P,j,k),mod2dense_get(P,i,k),bit);
	    
	    mod2dense_set(P,i,k, bit);
	  }

	  for (k=0; k<numNewSymbols; k++){
	    bit = (mod2dense_get(L,j,k) + mod2dense_get(L,i,k)) %2;
	    mod2dense_set(L,i,k, bit);
	  }

	  //exit(-1);	
	  NUMELIMONES ++;
	}
      }
      else{
	if( mod2dense_get(L,i,j)!=1){
	  goto ge_error;
	}
      }
    }
  }
  fprintf(stderr," makeExtCodeGenerator eliminates %d off-diagonal ones in L sub-matrix. ",NUMELIMONES);

  mod2dense_copy(P,G1);
  
  /* [Original routine:] */
  /* Invert L w possible re-ordered bits*/
  //mod2dense_invert_reorder(L,LI);
  //mod2dense_multiply(LI,P,G1);
  
  /* Count and print number of 1s. */

  c = 0;
  for (i = 0; i<numNewSymbols; i++)
    { for (j = 0; j<N0; j++)
      { c += mod2dense_get(G1,i,j);
      }
    }
  fprintf(stderr,
	  "Number of 1s per check in Inv(L) X P of extension generator matrix is %.1f\n", (double)c/numNewSymbols);
  
  /* Write the represention of the generator matrix to the file. */

  intio_write(f,M);
  intio_write(f,N);

  mod2dense_write (f, G1);

  for (i = 0; i<numNewSymbols; i++)
    intio_write(f,new_bit_index[i]);
    

  /* Free up the memory */

  mod2dense_free(DH);
  mod2dense_free(P);
  mod2dense_free(L);
  mod2dense_free(LI);
  mod2dense_free(G1);

  free(col1);
  free(row1);
  free(new_bit_index);

  return;
  
 ge_error:
  {fprintf(stderr,"GE subroutine does not find pivot element.\n");
  exit(-1);
  }

}


void makeBaseCodeGenerator(FILE *f){
 
  mod2dense *DH, *A, *A2, *AI, *B, *G1;
  int i, j, c, c2, n;
  int *rows_inv;
  //FILE *g;

  DH = mod2dense_allocate(M,N);
  AI = mod2dense_allocate(M,M);
  B  = mod2dense_allocate(M,N-M);
  G1  = mod2dense_allocate(M,N-M);

  mod2sparse_to_dense(H,DH);
  mod2dense_fliplr(DH);

  /* Invert using whatever selection of rows/columns is needed to get a non-singular sub-matrix. */

  A  = mod2dense_allocate(M,N);
  A2 = mod2dense_allocate(M,N);

  //int tt;
  //for(tt=0;  tt<10; tt++)
  //  fprintf(stderr,"cols[%d] = %d\n",tt+M,cols[tt+M]);

  n = mod2dense_invert_selected(DH,A2,rows,cols);
  
  //fprintf(stderr,"\n");
  //int tt;
  //for(tt=0;  tt<M; tt++)
  //  if( tt+M != cols[tt+M])
  //    fprintf(stderr,"cols[%d] = %d\n",tt+M,cols[tt+M]);

  //g = open_file_std("A2.out","wb");
  //mod2dense_print(g,A2);
  //fclose(g);

  mod2sparse_to_dense(H,DH);  /* DH was destroyed by invert_selected */
  mod2dense_fliplr(DH);

  if (n>0)
    { fprintf(stderr,"Note: Parity check matrix has %d redundant checks\n",n);
    }

  rows_inv = chk_alloc (M, sizeof *rows_inv);

  for (i = 0; i<M; i++)
    { rows_inv[rows[i]] = i;
    }

  mod2dense_copyrows(A2,A,rows);
  mod2dense_copycols(A,A2,cols);
  mod2dense_copycols(A2,AI,rows_inv);

  free(rows_inv);
  mod2dense_free(A);
  mod2dense_free(A2);

  mod2dense_copycols(DH,B,cols+M);

  /* Form final generator matrix. */
  
  mod2dense_multiply(AI,B,G1);

  /* Compute and print number of 1s. */

  c = 0;
  for (i = 0; i<M; i++)
    { for (j = 0; j<N-M; j++)
      { c += mod2dense_get(G1,i,j);
      }
    }
  fprintf(stderr,
	  "Number of 1s per check in Inv(A) X B of base generator matrix is %.1f\n", (double)c/M);
  
  /* Write the represention of the generator matrix to the file. */

  intio_write(f,M);
  intio_write(f,N);

  for (i = 0; i<N; i++) 
  { intio_write(f,cols[i]);
  }
  
  mod2dense_write (f, G1);

  mod2dense_free(DH);
  mod2dense_free(AI);
  mod2dense_free(B);
  mod2dense_free(G1);
}



/* PRINT USAGE MESSAGE AND EXIT. */

void usage(void)
{ fprintf (stderr, 
   "Usage:  make-gen numCodes baseCodeName\n");
  exit(1);
}
