/* Copyright 2007 Noah B. Jacobsen.
 * 
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

/******************************************************************************/
/* *******Includes code from:*********                                        */
/* Copyright (c) 2000, 2001 by Radford M. Neal                                */
/*                                                                            */
/* Permission is granted for anyone to copy, use, modify, or distribute this  */
/* program and accompanying programs and documents for any purpose, provided  */
/* this copyright notice is retained and prominently displayed, along with    */
/* a note saying that the original programs are available from Radford Neal's */
/* web page, and note is made of any changes made to the programs.            */
/******************************************************************************/
#include <stdio.h>
#include <stdlib.h>
#include <math.h>

#include "alloc.h"
#include "intio.h"
#include "open.h"
#include "mod2sparse.h"
#include "mod2dense.h"
#include "mod2convert.h"
#include "ncode.h"


/* VARIABLES DECLARED IN NCODE.H.  These global variables are set to
   representations of the parity check and generator matrices by read_pchk
   and read_gen. */

mod2sparse *H;		/* Parity check matrix */
int harqNum;            /* H-ARQ transmission number */

int M;			/* Number of rows in parity check matrix */
int N;			/* Number of columns in parity check matrix */

int *cols;		/* Ordering of columns in base generator matrix */
int *rows;		/* Ordering of rows in base generator matrix */

mod2dense **G;		/* Generator matrices:      */
int numCodes;           /*  1st harqNum             */

int *codeLength;        /*   entries are populated  */

/* READ PARITY CHECK MATRIX.  Sets the H, M, and N global variables.  If an
   error is encountered, a message is displayed on standard error, and the
   program is terminated. */

void read_pchk
( char *pchk_file
)
{
  FILE *f;

  f = open_file_std(pchk_file,"rb");
  if (f==NULL)
  { fprintf(stderr,"Can't open parity check file: %s\n",pchk_file);
    exit(1);
  }

  if (intio_read(f)!=('P'<<8)+0x80)
  { fprintf(stderr,"File %s doesn't contain a parity check matrix\n",pchk_file);
    exit(1);
  }


  H = mod2sparse_read(f);

  if (H==0)
  { fprintf(stderr,"Error reading parity check matrix from %s\n",pchk_file);
    exit(1);
  }

  M = mod2sparse_rows(H);
  N = mod2sparse_cols(H);

  fclose(f);
}


/* READ GENERATOR MATRIX. Sets all remaining global variables. */

void read_gen(char *gen_file){	 /* Name of generator matrix file */
  
  int M0, N0, M1, N1, K=N-M;
  FILE *f;
  int i,j;

  f = open_file_std(gen_file,"rb");
  if (f==NULL)
  { fprintf(stderr,"Can't open generator matrix file: %s\n",gen_file);
    exit(1);
  }

  if (intio_read(f)!=('G'<<8)+0x80)
  { fprintf(stderr,"File %s doesn't contain a generator matrix\n",gen_file);
    exit(1);
  }

  numCodes = intio_read(f);

  if(numCodes<0)
    { fprintf(stderr,"Number of codes negative\n");
    exit(1);
    }

  M1 = intio_read(f);
  N1 = intio_read(f);
  if (feof(f) || ferror(f)) goto error;

  if (K != N1-M1){ 
    fprintf(stderr, "Generator matrix and parity-check matrix are incompatible\n");
    exit(1);
    }
  
  cols = chk_alloc (N1, sizeof *cols);
  rows = chk_alloc (M1, sizeof *rows);

  for (i = 0; i<N1; i++)
  { cols[i] = intio_read(f);
    if (feof(f) || ferror(f)) goto error;
  }

  codeLength = chk_alloc (numCodes, sizeof *codeLength);
  codeLength[0] = N1;

  G = chk_alloc (numCodes, sizeof *G);

  if ((G[0] = mod2dense_read(f)) == 0) goto error;

  if (mod2dense_rows(G[0])!=M1 || mod2dense_cols(G[0])!=K) goto garbled;
  
  if (N==N1){
    harqNum = 0;
    goto success;
  }
  
  for (i=1; i<numCodes; i++){
    M0 = M1;
    N0 = N1;
    M1 = intio_read(f);
    N1 = intio_read(f);

    codeLength[i] = N1;

    if (feof(f) || ferror(f)) goto error;
    
    if (K != N1-M1){
      fprintf(stderr, "Generator matrix-%d and parity-check matrix are incompatible\n",i);
      exit(1);
    }

    if ((G[i] = mod2dense_read(f)) == 0) goto error;

    if (mod2dense_rows(G[i])!=(M1-M0) || mod2dense_cols(G[i])!=N0) goto garbled;

    if (N==N1){
      harqNum = i;
      goto success;
    }
  }
  
  fprintf(stderr, "Parity-matrix length is incompatible with the generator matrices\n",harqNum);
  exit(1);

 success:
  fclose(f);
  return;

 error:
  fprintf(stderr,"Error reading generator matrix from file %s\n",gen_file);
  exit(1);

 garbled:
  fprintf(stderr,"Garbled generator matrix in file %s\n",gen_file);
  exit(1);
}
