/* Copyright 2007 Noah B. Jacobsen.
 * 
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

/************************************************************************/
/*        *******Includes code from:*********                           */
/*        Free software: Progressive edge-growth (PEG) algorithm        */
/*        Created by Xiaoyu Hu                                          */
/*                   Evangelos Eletheriou                               */
/*                   Dieter Arnold                                      */
/*        IBM Research, Zurich Research Lab., Switzerland               */
/**********************************************************************/
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <iostream.h>
#include <iomanip.h>
#include <fstream.h>
#include <math.h>
#include "BigTriang.h"
#include "ExtTriang.h"
#include "Random.h"
#include "CyclesOfGraph.h"

#define EPS 1e-6
void readBaseCode(ifstream &, int, int &, int *, Random *, int &, int *, int &, int *, double *);
void initNextCode(ifstream &, int, int &, int *, Random *, BigTriang *, int &, int *, int &);

int main(int argc, char * argv[]){
  int i, n, N=0, K, M0;
  char codeName[100], degFileName[100], baseCodeName[100];
  int numCodes, *degSeq, numSplitParity;
  BigTriang *bigTriang;
  ExtTriang *extTriang;
  CyclesOfGraph *cog;
  Random *rand;
  int m, deg[50];
  int parityNumDeg, parityDeg[50];
  double parityDegFrac[50];

  int numArgs=(argc-1)/2;
  if (argc<7) {
  USE:
    cout<<"********************************************************************************"<<endl;
    cout<<" Usage Reminder: TriangPEG -numK K -codeName CodeName -degFileName DegFileName "<<endl;
    cout<<"********************************************************************************"<<endl;
    exit(-1);
  }else {
    for(i=0;i<numArgs;i++){
      if (strcmp(argv[2*i+1], "-numK")==0) {
	K=atoi(argv[2*i+2]);
      } else if(strcmp(argv[2*i+1], "-codeName")==0) {
	strcpy(baseCodeName, argv[2*i+2]); 
      } else if(strcmp(argv[2*i+1], "-degFileName")==0) {
	strcpy(degFileName, argv[2*i+2]); 
      } else{
    goto USE;
      }
    }
  }

  rand=new Random();  //(12345678l, 987654321lu);

  ifstream infn(degFileName);
  if (!infn) {cout << "\nCannot open file " << degFileName << endl; exit(-1); }

  //base code 
  infn >>numCodes;  
  degSeq=new int[11*K];

  readBaseCode(infn, K, N, degSeq, rand, m, deg, parityNumDeg, parityDeg, parityDegFrac); 

  if(numCodes>1)
    strcat(strcpy(codeName,baseCodeName),"0.pchk");
  else
    strcpy(codeName,baseCodeName);

  bigTriang=new BigTriang(N-K,N,degSeq,codeName,rand,parityNumDeg,parityDeg,parityDegFrac);
  (*bigTriang).writeToFile();

  for(n=1;n<numCodes;n++){
    //nth extension code
    initNextCode(infn,K,N,degSeq,rand,bigTriang,m,deg,numSplitParity); 
  
    sprintf(codeName,"%s%d.pchk",baseCodeName,n);
    extTriang=new ExtTriang(N-K,N,degSeq,numSplitParity,codeName,bigTriang,rand);  
  
    delete extTriang; extTriang=NULL;
    (*bigTriang).writeToFile();
  }

  delete bigTriang; bigTriang=NULL;
  infn.close();

  //computing local girth distribution  
  /*  if(N<10000) {
    cout<<" Now computing the local girth on the global Tanner graph setting. "<<endl;
    cout<<"     might take a bit long time. Please wait ...                   "<<endl;
    (*bigTriang).loadH();
    cog=new CyclesOfGraph(N-K, N, (*bigTriang).H);
    (*cog).getCyclesTable();
    (*cog).printCyclesTable();
    delete cog;
    cog=NULL;
    }*/

  delete rand;
  delete [] degSeq; degSeq=NULL;
}


void initNextCode(ifstream &infn, int K, int &N, int *degSeq, Random *rand, BigTriang *bigTriang, int &m, int *deg, int &numSplitParity){
  int i,j,n,N0,index;
  int *initDegSeq, *newDegSeq, *initSymbolIndex;
  double maxInitDeg, *degFrac, rate, a;
  int numFree, numAssigned, numCur;
  int numAdjustedSymbols, totAdjustedDiff;
  int numDegOne, numDegTwo, numKeepTwo;

  N0 = (*bigTriang).N;
  maxInitDeg=deg[m-1];
    
  //----Read in next code
  infn >>rate;
  //N = (int) round ( (double) K/rate );
  N = (int) floor ( (double) K/rate + 0.5);
  
  infn >>a;
  //numSplitParity = (int) round( (double) a*(N-K) );
  numSplitParity = (int) floor( (double) a*(N-K) + 0.5);

  infn >>m;
  degFrac=new double[m];

  for(i=0;i<m;i++) infn>>deg[i];
  for(i=0;i<m;i++) infn>>degFrac[i];

  //----Compute new degree sequence
  double dtmp=0.0;
  for(i=0;i<m;i++) dtmp+=degFrac[i];
  if(fabs(dtmp-1.0)>EPS) {
    cout.setf(ios::fixed, ios::floatfield);
    cout <<"\nTriangPeg::initNextCode ERROR invalid degree distribution: sum != 1.0 but "<<setprecision(10)<<dtmp<<endl; 
    exit(-1);
  } 

  for(i=1;i<m;i++) degFrac[i]+=degFrac[i-1];
  newDegSeq=new int[N];
  
  for(i=0;i<N;i++) {
    dtmp=(double)i/N;
    for(j=m-1;j>=0;j--) {
      if(dtmp>degFrac[j]) break;
    }
    if(dtmp<degFrac[0]) newDegSeq[i]=deg[0];
    else newDegSeq[i]=deg[j+1];
  }

  
  //----Compute sorted original degree sequence
  initDegSeq=new int[N];
  for(index=0;index<N-N0;index++) initDegSeq[index]=0;

  for(i=1;i<=maxInitDeg;i++){
    for(n=0;n<N0;n++){
      if( (*bigTriang).nodesInGraph[n].numOfConnectionSymbolBit==i ){
	initDegSeq[index]=i;
	index++;
      }
    }
  }
  if(index!=N){cout <<"\nTriangPeg::initNextCode ERROR index = "<<index<<", N = "<<N<<"\n"<<endl;exit(-1);}
  

  //----Check compatibility and modify newDegSeq if required
  numAdjustedSymbols=0;
  totAdjustedDiff=0;
  numAssigned=0;

  for(j=0;j<m;j++){
    numCur=0;
    numFree=0;
    
    for(n=numAssigned;n<N;n++){
      if(newDegSeq[n]==deg[j])
	numCur++;
      if(initDegSeq[n]<=deg[j])
	numFree++;
    }
    
    if(numCur>=numFree)
      for(i=numFree;i<numCur;i++){
	newDegSeq[numAssigned+i]=initDegSeq[numAssigned+i];
	
	totAdjustedDiff += (initDegSeq[numAssigned+i]-deg[j]);
	numAdjustedSymbols++;
      }
    numAssigned+=numCur;
  }
  if(numAssigned!=N){cout <<"\nTriangPeg::initNextCode ERROR numAssigned = "<<numAssigned<<", N = "<<N<<"\n"<<endl;exit(-1);}
  if(numAdjustedSymbols>0)
    cout<<"\nTriangPeg::initNextCode adjusted "<<numAdjustedSymbols<<" columns for total "<<totAdjustedDiff<<" edges"<<endl;
  

  //----Special handlement of degree-1 and degree-2 nodes
  numDegOne=0;
  numDegTwo=0;
  for(n=0;n<N;n++){
    if(newDegSeq[n]==1)
      numDegOne++;
    if(newDegSeq[n]==2)
      numDegTwo++;
  }
  //Force degree-one nodes
  if( numDegOne==0){

    int ForceDegOne = 100;
    if( numDegTwo>ForceDegOne){
      numDegOne=ForceDegOne;
      numDegTwo-=ForceDegOne;
      
      for(n=0;n<ForceDegOne;n++)
	newDegSeq[n]=1;
    }
    else{
      numDegOne=1;
      numDegTwo-=1;
      newDegSeq[0]=1;
    }
  }

  //----Index new degree sequence 
  initSymbolIndex=new int[N];

  for(index=0;index<numDegOne;index++)
    initSymbolIndex[index]=N-index-1;
  
  numKeepTwo=0;
  for(n=N0-1;n>=0;n--)
    if( (*bigTriang).nodesInGraph[n].keepAsDegreeTwo==1 ){    
      initSymbolIndex[index]=n;
      numKeepTwo++;
      index++;
    }

  if(index+numSplitParity>numDegTwo+numDegOne){
    cout <<"\nTriangPeg::initNextCode WARNING numSplitParity is reduced by "<<numSplitParity-numDegOne-numDegTwo+index<<" due to not enough degree two nodes\n"<<endl;
    numSplitParity=numDegTwo+numDegOne-index;
  }

  if(numDegOne+numSplitParity>N-N0){
    cout <<"\nTriangPeg::initNextCode WARNING numSplitParity is reduced by "<<numSplitParity-(N-N0-numDegOne)<<" due to a large number of degree one nodes\n"<<endl;
    numSplitParity=N-N0-numDegOne;
  }

  for(n=N0+numSplitParity-1;n>=N0;n--){
    initSymbolIndex[index]=n;
    index++;
  }

  for(n=N-numDegOne-1;n>=N0+numSplitParity;n--){
    initSymbolIndex[index]=n;
    index++;
  }
  if(index!=N-N0+numKeepTwo){cout <<"\nTriangPeg::initNextCode ERROR index hosed\n"<<endl;exit(-1);}

  for(i=1;i<=maxInitDeg;i++){
    for(n=N0-1;n>=0;n--)
      if( (*bigTriang).nodesInGraph[n].keepAsDegreeTwo==0 ){
	if( (*bigTriang).nodesInGraph[n].numOfConnectionSymbolBit==i ){
	  initSymbolIndex[index] = n;
	  index++;
	}
      }
  }
  if(index!=N){cout <<"\nTriangPeg::initNextCode ERROR index = "<<index<<", N = "<<N<<"\n"<<endl;exit(-1);}
  
  //Put back in order
  for(n=0;n<N;n++)
    degSeq[initSymbolIndex[n]]=newDegSeq[n];
  
  for(n=0;n<N0;n++)
    if(degSeq[n]<(*bigTriang).nodesInGraph[n].numOfConnectionSymbolBit){
      cout <<"\nTriangPeg::initNextCode ERROR n=" <<n<<" and degSeq ("<<degSeq[n]<<") is < than numOfConnectionSymbolBit ("<<(*bigTriang).nodesInGraph[n].numOfConnectionSymbolBit<<")\n"<< endl; exit(-1);
    }

  delete [] degFrac; degFrac=NULL;
  delete [] newDegSeq; newDegSeq=NULL;
  delete [] initDegSeq; initDegSeq=NULL;
  delete [] initSymbolIndex; initSymbolIndex=NULL;
  
  return;
}


void readBaseCode(ifstream &infn, int K, int &N, int *degSeq, Random *rand, int &m, int *deg, int &parityNumDeg, int *parityDeg, double *parityDegFrac){
  int i,j;
  double *degFrac, rate;

  infn >>rate;
  //N =(int) round ((double) K/rate );
  N =(int) floor ((double) K/rate + 0.5 );
  
  infn >>parityNumDeg;
  for(i=0;i<parityNumDeg;i++) infn>>parityDeg[i];
  for(i=0;i<parityNumDeg;i++) infn>>parityDegFrac[i];
  
  infn >>m;
  degFrac=new double[m];

  for(i=0;i<m;i++) infn>>deg[i];
  for(i=0;i<m;i++) infn>>degFrac[i];

  double dtmp=0.0;
  for(i=0;i<m;i++) dtmp+=degFrac[i];
  if(fabs(dtmp-1.0)>EPS) {
    cout.setf(ios::fixed, ios::floatfield);
    cout <<"\n Invalid degree distribution (node perspective): sum != 1.0 but "<<setprecision(10)<<dtmp<<endl; exit(-1); 
  } 

  for(i=1;i<m;i++) degFrac[i]+=degFrac[i-1];

  for(i=0;i<N;i++) {
    dtmp=(double)i/N;
    for(j=m-1;j>=0;j--) {
      if(dtmp>degFrac[j]) break;
    }
    if(dtmp<degFrac[0]) degSeq[N-i-1]=deg[0];
    else degSeq[N-i-1]=deg[j+1];
  }

  delete [] degFrac; degFrac=NULL;
  return;
}
